require "Apollo"
require "AbilityBook" 
require "ActionSetLib"
require "GameLib"
require "Unit"

local keProcType = {
	Critical = 1,
	Dodge = 2,
	Buff = 3
}

local karProcSettings = { 
	-- Engineer
	{ strName = "Quick Burst", nProcType = keProcType.Critical, nDuration = 5 },
	{ strName = "Feedback", nProcType = keProcType.Dodge, nDuration = 5 },
	-- Medic
	{ strName = "Atomize", nProcType = keProcType.Buff, strBuffName = "Clear!", nDuration = 5 },
	{ strName = "Dual Shock", nProcType = keProcType.Buff, strBuffName = "Clear!", nDuration = 5 },
	-- Spellslinger
	{ strName = "Flame Burst", nProcType = keProcType.Critical, nDuration = 5 },
	-- Stalker
	{ strName = "Punish", nProcType = keProcType.Critical, nDuration = 5 },
	{ strName = "Decimate", nProcType = keProcType.Dodge, nDuration = 5 },
	-- Warrior
	{ strName = "Breaching Strikes", nProcType = keProcType.Critical, nDuration = 5 },
	{ strName = "Atomic Spear", nProcType = keProcType.Dodge, nDuration = 5 }
}

local AlpacaGameLibrary = {}

AlpacaGameLibrary.EnumUnitType = {
	Engineer = 1,
  Esper = 2,
  Medic = 3,
  Spellslinger = 4,
  Stalker = 5,
  Warrior = 6,
	Pet = 7,
  NonPlayerFriendly = 8,
  NonPlayerNeutral = 9,
  NonPlayerHostile = 10,
  Simple = 12,
	Unknown = 13
}

AlpacaGameLibrary.EnumUnitFlags = {
  PlayerFriendly = 1,
  PlayerFriendlyPvp = 2,
  PlayerHostile = 3,
  PlayerHostilePvp = 4,
  PlayerGroupMember = 5,
  PlayerGroupMemberPvp = 6,
  PetFriendly = 7,
  PetFriendlyPvp = 8,
  PetHostile = 9,
  PetHostilePvp = 10,
  NonPlayerFriendly = 11,
  NonPlayerHostile = 12,
  NonPlayerNeutral = 13,
  Simple = 14,
  Unknown = 15
}

AlpacaGameLibrary.EnumUnitDifficulty = {
	Trivial = 1,
	Inferior = 2,
	Minor = 3,
	Easy = 4,
	Average = 5,
	Moderate = 6,
	Tough = 7,
	Hard = 8,
	Impossible = 9  
}

function AlpacaGameLibrary:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self 
  return o
end

function AlpacaGameLibrary:Init()
  self.tDisableTimers = {}
  
  Apollo.RegisterPackage(self, "AlpacaGameLibrary", 1, {})
end

function AlpacaGameLibrary:GetInnateSpells()
  local arSpells = {}
  
  for i, objSpell in ipairs(GameLib.GetClassInnateAbilitySpells().tSpells) do
    if i % 2 == 1 then
      table.insert(arSpells, objSpell)
    end
  end
  
  return arSpells
end

function AlpacaGameLibrary:SetInnate(nSpellId)
  local arSpells = self:GetInnateSpells()
  
  for i, objSpell in ipairs(arSpells) do
    if objSpell:GetId() == nSpellId then
      GameLib.SetCurrentClassInnateAbilityIndex(i)
      break
    end
  end
end

function AlpacaGameLibrary:GetMountSpells()
  local arMounts = {} 
  
 	for _, tMount in ipairs(AbilityBook.GetAbilitiesList(Spell.CodeEnumSpellTag.Mount) or {}) do
		table.insert(arMounts, tMount.tTiers[1].splObject)
  end
  
  return arMounts
end

function AlpacaGameLibrary:SetMount(nSpellId)
  GameLib.SetShortcutMount(nSpellId)
end

function AlpacaGameLibrary:GetPotionItems()
  local unitPlayer = GameLib.GetPlayerUnit()
	local arItems = unitPlayer and unitPlayer:IsValid() and unitPlayer:GetInventoryItems() or {}
  local arPotions = {}
  
  for _, tItem in ipairs(arItems) do
    if tItem and tItem.itemInBag and tItem.itemInBag:GetItemCategory() == 48 then
      table.insert(arPotions, tItem.itemInBag)
    end
  end
  
  return arPotions
end
	
function AlpacaGameLibrary:SetPotion(nItemId)
  GameLib.SetShortcutPotion(nItemId)
end

function AlpacaGameLibrary:GetClassProcs()
  local fnGetAbility = function(strName)
    for _, tAbility in ipairs(AbilityBook.GetAbilitiesList()) do
      if tAbility.strName == strName then
        return tAbility
      end
    end
  end

  local arClassProcs = {}

  for _, tProcSettings in ipairs(karProcSettings) do
    local tAbility = fnGetAbility(tProcSettings.strName)
    if tAbility then
      tAbility.nProcType = tProcSettings.nProcType
      tAbility.strBuffName = tProcSettings.strProcName
      tAbility.nDuration = tProcSettings.nDuration
      table.insert(arClassProcs, tAbility)
    end
  end

  return arClassProcs
end

function AlpacaGameLibrary:GetActionBarProcs()
  local tActiveAbilityIds = {}
  for _, nId in ipairs(ActionSetLib.GetCurrentActionSet()) do
    tActiveAbilityIds[nId] = true
  end
  
  local arClassProcs = self:GetClassProcs()
  local arActionBarProcs = {}
  
  for _, tAbility in ipairs(arClassProcs) do
    if tActiveAbilityIds[tAbility.nId] and tAbility.nCurrentTier and tAbility.nCurrentTier > 0 then 
      table.insert(arActionBarProcs, tAbility)        
    end
  end
  
  return arActionBarProcs
end

function AlpacaGameLibrary:GetProcSettings(strName)
	for _, tProcSettings in ipairs(karProcSettings) do
		if tProcSettings.strName == strName then
			return tProcSettings
		end
	end		
end

function AlpacaGameLibrary:GetProcTimes(tAbility, fLastCritical, fLastDodge)
  local tProcSettings = self:GetProcSettings(tAbility.strName)
	
	if not tProcSettings then return end
	
	local fProcRemaining = nil
	local fProcDuration = tProcSettings.nDuration
	local tSpell = tAbility.tTiers[tAbility.nCurrentTier].splObject
	local fCooldownRemaining = tSpell:GetCooldownRemaining()	
		
	if tProcSettings.nProcType == keProcType.Buff then 
		local tBuff = self:GetBuff(self:GetPlayerUnit(), tProcSettings.strBuffName)
		fProcRemaining = tBuff and tBuff.fTimeRemaining
	elseif tProcSettings.nProcType == keProcType.Critical then
		fProcRemaining = fLastCritical and math.max(0, fProcDuration - (os.clock() - fLastCritical))
	elseif tProcSettings.nProcType == keProcType.Dodge then
		fProcRemaining = fLastDodge and math.max(0, fProcDuration - (os.clock() - fLastDodge))
	end

	fProcRemaining = (fProcRemaining and fProcRemaining > 0) and fProcRemaining
	fCooldownRemaining = (fCooldownRemaining and fCooldownRemaining > 0) and fCooldownRemaining
	
	return fProcRemaining, fProcDuration, fCooldownRemaining
end

function AlpacaGameLibrary:GetBuff(unit, strName)
  local arBuffs = unit:GetBuffs().arBeneficial
    
  for _, tBuff in pairs(arBuffs) do
    if tBuff.splEffect:GetName() == strName then
      return tBuff
    end
  end
end

function AlpacaGameLibrary:GetResourceId(unit)
  local nClassId = unit:GetClassId()
  local nResourceId = nil
  
  if nClassId == GameLib.CodeEnumClass.Spellslinger then
    nResourceId = 4
  elseif nClassId == GameLib.CodeEnumClass.Stalker then
    nResourceId = 3  
  else
    nResourceId = 1
  end
  
  return nResourceId 
end

function AlpacaGameLibrary:IsInCombatOrDamaged(unit)
   local inCombat = unit:IsInCombat()
  
  if inCombat then
    return true
  end
  
  local lowShield = (unit:GetShieldCapacityMax() or 0) > 0 and unit:GetShieldCapacity() < unit:GetShieldCapacityMax() 
  local lowHealth = (unit:GetMaxHealth() or 0) > 0 and unit:GetHealth() < unit:GetMaxHealth()
  
  return lowShield or lowHealth
end

function AlpacaGameLibrary:GetUnitHealthPercent(unit)
  local nHealth = unit:GetHealth()
  local nMaxHealth = unit:GetMaxHealth()
  
  if nHealth and nHealth > 0 then
    return nHealth / nMaxHealth
  else
    return 0
  end
end

function AlpacaGameLibrary:GetUnitShieldPercent(unit)
  local nShield = unit:GetShieldCapacity()
  local nMaxShield = unit:GetShieldCapacityMax()

  if nShield and nShield > 0 then
    return nShield / nMaxShield
  else
    return 0
  end
end

function AlpacaGameLibrary:GetUnitAbsorbPercent(unit)
  local nAbsorb = unit:GetAbsorptionValue()
  local nMaxAbsorb = unit:GetAbsorptionMax()
    
  if nAbsorb and nAbsorb > 0 then
    return nAbsorb / nMaxAbsorb
  else
    return 0
  end
end

function AlpacaGameLibrary:GetUnitFocusPercent(unit) 
  local nFocus = unit:GetMana()
  local nMaxFocus = unit:GetMaxMana()
    
  if nFocus and nFocus > 0 then
    return nFocus / nMaxFocus
  else
    return 0
  end  
end

function AlpacaGameLibrary:GetUnitResourcePercent(unit)
  local nResourceId = AlpacaGameLibrary:GetResourceId(unit)
  local nResource = unit:GetResource(nResourceId)
  local nMaxResource = unit:GetMaxResource(nResourceId)

  if nResource and nResource > 0 then
    return nResource / nMaxResource
  else
    return 0
  end
end

function AlpacaGameLibrary:GetUnitSprintPercent(unit)
	local nSprint = unit:GetResource(0)
	local nMaxSprint = unit:GetMaxResource(0)
  
  if nSprint and nSprint > 0 then
    return nSprint / nMaxSprint
  else
    return 0
  end
end

function AlpacaGameLibrary:GetUnitDashPercent(unit)
  local nDash = unit:GetResource(7)
  local nMaxDash = unit:GetMaxResource(7)
  
  if nDash and nDash > 0 then
    return nDash / nMaxDash
  else
    return 0
  end  
end

function AlpacaGameLibrary:GetUnitCastPercent(unit)
  local bShouldShow = unit:ShouldShowCastBar()
  local nDuration = unit:GetCastDuration()
  local nElapsed = unit:GetCastElapsed()
  
  if bShouldShow and nElapsed and nElapsed > 0 then
    return nElapsed / nDuration
  else
    return 0
  end
end

function AlpacaGameLibrary:GetUnitDisablePercent(unit, arDisables)
  arDisables = arDisables or {
    Unit.CodeEnumCCState.Blind,
    Unit.CodeEnumCCState.Disarm,
    Unit.CodeEnumCCState.Fear,
    Unit.CodeEnumCCState.Knockdown,
    Unit.CodeEnumCCState.Stun,  
    Unit.CodeEnumCCState.Subdue,
    Unit.CodeEnumCCState.Tether
  }

  local fNow = os.clock()
  
  -- clear expired timers
  for nUnitId, tTimer in pairs(self.tDisableTimers) do
    if tTimer.fEnd < fNow then
      self.tDisableTimers[nUnitId] = nil
    end
  end
  
  -- get max disable time
  local fRemaining = 0
  
  for _, eDisable in ipairs(arDisables) do
    fRemaining = math.max(fRemaining, unit:GetCCStateTimeRemaining(eDisable))    
  end
  
  -- update timer
  local nUnitId = unit:GetId()

  if not fRemaining or fRemaining == 0 then
    self.tDisableTimers[nUnitId] = nil    
    return 0
  end
  
  if not self.tDisableTimers[nUnitId] then
    self.tDisableTimers[nUnitId] = { fStart = fNow, fEnd = fNow + fRemaining }
  elseif self.tDisableTimers[nUnitId].fEnd < fNow + fRemaining then
    self.tDisableTimers[nUnitId].fEnd = fNow + fRemaining
  end

  -- return percent
  local fPercent = fRemaining / (self.tDisableTimers[nUnitId].fEnd - self.tDisableTimers[nUnitId].fStart)
  
  return fPercent
end

function AlpacaGameLibrary:GetUnitType(unitPlayer, unitOther)  
  local eUnitType = nil
  local strUnitType = unitOther:GetType()
  
  if strUnitType == "Player" then
    local tLookupPlayerUnitType = {
      [GameLib.CodeEnumClass.Engineer]      = self.EnumUnitType.Engineer,
      [GameLib.CodeEnumClass.Esper]         = self.EnumUnitType.Esper,
      [GameLib.CodeEnumClass.Medic]         = self.EnumUnitType.Medic,
      [GameLib.CodeEnumClass.Spellslinger]  = self.EnumUnitType.Spellslinger,
      [GameLib.CodeEnumClass.Stalker]       = self.EnumUnitType.Stalker,    
      [GameLib.CodeEnumClass.Warrior]       = self.EnumUnitType.Warrior
    }

    eUnitType = tLookupPlayerUnitType[unitOther:GetClassId()]
  elseif strUnitType == "NonPlayer" then
    local tLookupNonPlayerUnitType = {
      [Unit.CodeEnumDisposition.Friendly] 	= self.EnumUnitType.NonPlayerFriendly,
      [Unit.CodeEnumDisposition.Hostile] 		= self.EnumUnitType.NonPlayerHostile,
      [Unit.CodeEnumDisposition.Neutral] 		= self.EnumUnitType.NonPlayerNeutral,
      [Unit.CodeEnumDisposition.Unknown] 		= self.EnumUnitType.NonPlayerNeutral
    } 
    
    eUnitType = tLookupNonPlayerUnitType[unitOther:GetDispositionTo(unitPlayer)]    
  else
    local tOtherUnitType = {
      ["Pet"] 		         = self.EnumUnitType.Pet,
      ["Esper Pet"]	       = self.EnumUnitType.Pet,
      ["Simple"] 		       = self.EnumUnitType.Simple,
      ["SimpleCollidable"] = self.EnumUnitType.Simple
    }

    eUnitType = tOtherUnitType[strUnitType]
  end

  if eUnitType then
    return eUnitType
  end

  return self.EnumUnitType.Unknown
end

function AlpacaGameLibrary:GetUnitFlags(unitPlayer, unitOther)
	local unitController = unitOther:GetUnitOwner() or unitOther
    
  local strUnitType = unitOther:GetType()
  
  if strUnitType ~= "Pet" and strUnitType ~= "Esper Pet" then
    strUnitType = unitController:GetType()
  end
  
	local eDisposition = unitController:GetDispositionTo(unitPlayer)
  local bIsPvpFlagged = unitController:IsPvpFlagged()
    
	if strUnitType == "Simple" or strUnitType == "SimpleCollidable" then
    return self.EnumUnitFlags.Simple
  elseif strUnitType == "Player" then     
		if eDisposition == Unit.CodeEnumDisposition.Friendly then
			if bIsPvpFlagged then
				return self.EnumUnitFlags.PlayerFriendlyPvp
			else
				return self.EnumUnitFlags.PlayerFriendly
			end	
		else
			if bIsPvpFlagged then
        return self.EnumUnitFlags.PlayerHostilePvp
      else
        return self.EnumUnitFlags.PlayerHostile
			end
		end
  elseif strUnitType == "Pet" or strUnitType == "Esper Pet" then
    if eDisposition == Unit.CodeEnumDisposition.Friendly then
			if bIsPvpFlagged then
				return self.EnumUnitFlags.PetFriendlyPvp
			else
				return self.EnumUnitFlags.PetFriendly
			end	
		else
			if bIsPvpFlagged then
        return self.EnumUnitFlags.PetHostilePvp
      else
        return self.EnumUnitFlags.PetHostile
			end
		end    
  elseif strUnitType == "NonPlayer" then
    if eDisposition == Unit.CodeEnumDisposition.Friendly then
      return self.EnumUnitFlags.NonPlayerFriendly
    elseif eDisposition == Unit.CodeEnumDisposition.Hostile then
      return self.EnumUnitFlags.NonPlayerHostile
    else
      return self.EnumUnitFlags.NonPlayerNeutral
    end
	end	
  
  return self.EnumUnitFlags.Unknown
end

function AlpacaGameLibrary:GetUnitDifficulty(unitPlayer, unitOther)
	if not self.nMinUnitDifficulty or not self.nMaxUnitDifficulty or not self.nAvgUnitDifficulty then
		self.nMinUnitDifficulty = 1
		self.nMaxUnitDifficulty = 1
		self.nAvgUnitDifficulty = 1
		
		for _, nValue in pairs(self.EnumUnitDifficulty) do
			self.nMinUnitDifficulty = math.min(self.nMinUnitDifficulty, nValue)
			self.nMaxUnitDifficulty = math.max(self.nMaxUnitDifficulty, nValue)
		end

		self.nAvgUnitDifficulty = math.ceil(self.nMaxUnitDifficulty / 2)
	end

	local nLevelDiff = unitPlayer:GetLevelDifferential(unitOther) or 0
				
	return math.min(self.nMaxUnitDifficulty, math.max(self.nMinUnitDifficulty, nLevelDiff + self.nAvgUnitDifficulty))
end

function AlpacaGameLibrary:GetPlayerUnit()
  local unit = GameLib.GetPlayerUnit()

  if unit:GetClassId() then
    return unit
  end
end

local AlpacaGameLibraryInst = AlpacaGameLibrary:new()
AlpacaGameLibraryInst:Init()
