require "Apollo"
require "ApolloColor"
require "Window"
require "XmlDoc"

local AlpacaStandardLibrary = nil
local AlpacaWindowLibrary = nil

local AlpacaColorPicker = {} 

-- Init

function AlpacaColorPicker:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
  return o
end

function AlpacaColorPicker:Init()
  local bHasConfigureFunction = false
  local strConfigureButtonText = nil
  local tDependencies = { "AlpacaStandardLibrary", "AlpacaWindowLibrary" }
  
  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

-- Load

function AlpacaColorPicker:OnLoad()
  Apollo.LoadSprites("AlpacaSprites.xml")
  
  AlpacaStandardLibrary = Apollo.GetPackage("AlpacaStandardLibrary").tPackage
  AlpacaWindowLibrary = Apollo.GetPackage("AlpacaWindowLibrary").tPackage
    
  self.xmlDoc = XmlDoc.CreateFromFile("AlpacaColorPicker.xml")
  self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

function AlpacaColorPicker:OnDocLoaded()
  if self.xmlDoc == nil or not self.xmlDoc:IsLoaded() then return end

  self:CreateColorPicker()
  
  Apollo.RegisterEventHandler("OpenAlpacaColorPicker", "OnOpenAlpacaColorPicker", self)
end

-- Events

function AlpacaColorPicker:OnOpenAlpacaColorPicker(objPlugin, arSettingPath, tColor)
  self:OpenColorPicker(objPlugin, arSettingPath, AlpacaStandardLibrary:Copy(tColor))
end

-- Conversions

function AlpacaColorPicker:HSVtoRGB(fHue, fSat, fVal)
  if fSat <= 0 then
    return fVal, fVal, fVal
  end

  if fHue >= 360 then 
    fHue = 0
  end
  
  local nSector = math.floor(fHue / 60)    
  local fFactorial = fHue / 60 - nSector
  
  local p = fVal * (1 - fSat)
  local q = fVal * (1 - (fSat * fFactorial))
  local t = fVal * (1 - (fSat * (1 - fFactorial)))
  
  if nSector == 0 then
    return fVal, t, p
  elseif nSector == 1 then
    return q, fVal, p
  elseif nSector == 2 then
    return p, fVal, t
  elseif nSector == 3 then
    return p, q, fVal
  elseif nSector == 4 then
    return t, p, fVal
  else
    return fVal, p, q
  end
end

function AlpacaColorPicker:RGBtoHSV(fRed, fGreen, fBlue)
  local fMin = math.min(fRed, fGreen, fBlue)
  local fMax = math.max(fRed, fGreen, fBlue)
  local fDelta = fMax - fMin

  if fMax == 0 then  
    return nil, 0, 0
  end

  local fHue = nil
  local fSat = fDelta / fMax  
  local fVal = fMax
  
  if fDelta > 0 then
    if fRed == fMax then
      fHue = (fGreen - fBlue) / fDelta
    elseif fGreen == fMax then
      fHue = 2 + (fBlue - fRed) / fDelta
    else
      fHue = 4 + (fRed - fGreen) / fDelta
    end
    
    fHue = fHue * 60
    
    if fHue < 0 then
      fHue = fHue + 360
    end
  end
    
  return fHue, fSat, fVal
end

-- ColorPicker

function AlpacaColorPicker:CreateColorPicker()
  self.wndColorPicker = Apollo.LoadForm(self.xmlDoc, "Form", nil, self)
  self.wndColorPicker:Show(false, true)
  
  self:ColorPickerChangeColor(AlpacaWindowLibrary:AlpacaColor(1, 1, 1, 1))
end

function AlpacaColorPicker:OpenColorPicker(objPlugin, arSettingPath, crColor)
  local tData = { objPlugin = objPlugin, arSettingPath = arSettingPath }

  self:ColorPickerChangeColor(crColor)

  self.wndColorPicker:SetData(tData)
  self.wndColorPicker:Show(true)
  self.wndColorPicker:ToFront()
end

function AlpacaColorPicker:ColorPickerReadFromMarkers()
  local _, fRangeVerticalOffset = self.wndColorPicker:FindChild("Range:Marker"):GetAnchorPoints()
  local fGradientHorizontalOffset, fGradientVerticalOffset = self.wndColorPicker:FindChild("Gradient:Marker"):GetAnchorPoints()
  local fAlphaHorizontalOffset, _ = self.wndColorPicker:FindChild("Alpha:Marker"):GetAnchorPoints()
  
  local fHue = math.max(0, math.min(1, 1 - fRangeVerticalOffset)) * 360
  local fSat = math.max(0, math.min(1, fGradientHorizontalOffset))
  local fVal = (1 - math.max(0, math.min(1, fGradientVerticalOffset)))
  local fAlpha = math.max(0, math.min(1, fAlphaHorizontalOffset))
    
  local fRed, fGreen, fBlue = self:HSVtoRGB(fHue, fSat, fVal)
  
  return AlpacaWindowLibrary:AlpacaColor(fRed, fGreen, fBlue, fAlpha)
end

function AlpacaColorPicker:ColorPickerReadFromNumbers()
  local wndRedValue = self.wndColorPicker:FindChild("RedValue")
  local wndGreenValue = self.wndColorPicker:FindChild("GreenValue")
  local wndBlueValue = self.wndColorPicker:FindChild("BlueValue")
  local wndAlphaValue = self.wndColorPicker:FindChild("AlphaValue")
  
  local fRed = (tonumber(wndRedValue:GetText()) or 0) / 255
  local fGreen = (tonumber(wndGreenValue:GetText()) or 0) / 255
  local fBlue = (tonumber(wndBlueValue:GetText()) or 0) / 255
  local fAlpha = tonumber(wndAlphaValue:GetText())
  
  return AlpacaWindowLibrary:AlpacaColor(fRed, fGreen, fBlue, fAlpha)
end

function AlpacaColorPicker:ColorPickerReadFromPreview()
  local wndPreview = self.wndColorPicker:FindChild("Preview")
  
  return AlpacaWindowLibrary:ToAlpacaColor(wndPreview:GetBGColor())
end

function AlpacaColorPicker:ColorPickerChangeColor(crColor, wndControl)
  local wndGradientSprite = self.wndColorPicker:FindChild("Gradient:Sprite")
  local wndRangeSprite = self.wndColorPicker:FindChild("Range:Sprite")
  local wndAlphaSprite = self.wndColorPicker:FindChild("Alpha:Sprite")
  local wndRedValue = self.wndColorPicker:FindChild("RedValue")
  local wndGreenValue = self.wndColorPicker:FindChild("GreenValue")
  local wndBlueValue = self.wndColorPicker:FindChild("BlueValue")
  local wndHexValue = self.wndColorPicker:FindChild("HexValue")  
  local wndAlphaValue = self.wndColorPicker:FindChild("AlphaValue")  

  local fRed = 1
  local fGreen = 1
  local fBlue = 1
  local fAlpha = 1
  
  if crColor then
    fRed = crColor.fRed
    fGreen = crColor.fGreen
    fBlue = crColor.fBlue
    fAlpha = crColor.fAlpha
  end
  
  local nRed = AlpacaStandardLibrary:Round(fRed * 255, 0)
  local nGreen = AlpacaStandardLibrary:Round(fGreen * 255, 0)
  local nBlue = AlpacaStandardLibrary:Round(fBlue * 255, 0)
  local fHue, fSat, fVal = self:RGBtoHSV(fRed, fGreen, fBlue)
  local strHex = string.format('%02x%02x%02x', nRed, nGreen, nBlue)
  
  if not fHue then
    local _, fRangeVerticalOffset = self.wndColorPicker:FindChild("Range:Marker"):GetAnchorPoints()  
    fHue = math.max(0, math.min(1, 1 - fRangeVerticalOffset)) * 360
  end

  if wndControl ~= wndGradientSprite then
    local fGradientRed, fGradientGreen, fGradientBlue = self:HSVtoRGB(fHue, 1, 1)
    local fHorizontalOffset = fSat 
    local fVerticalOffset = 1 - fVal
        
    self.wndColorPicker:FindChild("Gradient:Marker"):SetAnchorPoints(fHorizontalOffset, fVerticalOffset, fHorizontalOffset, fVerticalOffset)
    wndGradientSprite:SetBGColor(ApolloColor.new(fGradientRed, fGradientGreen, fGradientBlue, 1))
  end
  
  if wndControl ~= wndRangeSprite then  
    local fVerticalOffset = 1 - fHue / 360
    
    if not fVerticalOffset or fVerticalOffset < 0 or fVerticalOffset > 1 then
      fVerticalOffset = 1
    end
    
    self.wndColorPicker:FindChild("Range:Marker"):SetAnchorPoints(0, fVerticalOffset, 1, fVerticalOffset)
  end
  
  if wndControl ~= wndAlphaSprite then  
    self.wndColorPicker:FindChild("Alpha:Marker"):SetAnchorPoints(fAlpha, 0, fAlpha, 1)
  end
  
  if wndControl ~= wndRedValue then
    wndRedValue:SetText(nRed)
  end
  
  if wndControl ~= wndGreenValue then
    wndGreenValue:SetText(nGreen)
  end
  
  if wndControl ~= wndBlueValue then
    wndBlueValue:SetText(nBlue)
  end
  
  if wndControl ~= wndHexValue then
    wndHexValue:SetText(strHex)
  end
  
  if wndControl ~= wndAlphaValue then
    wndAlphaValue:SetText(AlpacaStandardLibrary:Round(fAlpha, 5))
  end
  
  self.wndColorPicker:FindChild("Alpha:Sprite"):SetBGColor(AlpacaWindowLibrary:ToApolloColor(crColor))
  self.wndColorPicker:FindChild("Preview"):SetBGColor(AlpacaWindowLibrary:ToApolloColor(crColor))
end

-- ColorPicker:Range

function AlpacaColorPicker:OnColorPickerRangeStart(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
  if wndHandler == wndControl then
    self.bColorPickerRangeSelected = true
    self:ColorPickerMoveRangeMarker(nLastRelativeMouseX, nLastRelativeMouseY)
  end
end

function AlpacaColorPicker:OnColorPickerRangeStop(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation)
  if wndHandler == wndControl then
    self.bColorPickerRangeSelected = false
  end
end

function AlpacaColorPicker:OnColorPickerRangeMove(wndHandler, wndControl, nLastRelativeMouseX, nLastRelativeMouseY)
  if wndHandler == wndControl then
    if self.bColorPickerRangeSelected then
      self:ColorPickerMoveRangeMarker(nLastRelativeMouseX, nLastRelativeMouseY)
    end
  end
end

function AlpacaColorPicker:ColorPickerMoveRangeMarker(nLastRelativeMouseX, nLastRelativeMouseY)
  local wndRange = self.wndColorPicker:FindChild("Range")
  local wndSprite = wndRange:FindChild("Sprite")
  local wndMarker = wndRange:FindChild("Marker")
  local nMarkerPoint = math.max(0, math.min(1, nLastRelativeMouseY / wndSprite:GetHeight()))

  wndMarker:SetAnchorPoints(0, nMarkerPoint, 1, nMarkerPoint)
  self:ColorPickerChangeColor(self:ColorPickerReadFromMarkers())
end

-- ColorPicker:Gradient

function AlpacaColorPicker:OnColorPickerGradientStart(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation)
  if wndHandler == wndControl then
    self.bColorPickerGradientSelected = true
    self:ColorPickerMoveGradientMarker(nLastRelativeMouseX, nLastRelativeMouseY)
  end
end

function AlpacaColorPicker:OnColorPickerGradientStop(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
  if wndHandler == wndControl then
    self.bColorPickerGradientSelected = false
  end
end

function AlpacaColorPicker:OnColorPickerGradientMove(wndHandler, wndControl, nLastRelativeMouseX, nLastRelativeMouseY)
  if wndHandler == wndControl then
    if self.bColorPickerGradientSelected then
      self:ColorPickerMoveGradientMarker(nLastRelativeMouseX, nLastRelativeMouseY)
    end
  end
end

function AlpacaColorPicker:ColorPickerMoveGradientMarker(nLastRelativeMouseX, nLastRelativeMouseY)
  local wndGradient = self.wndColorPicker:FindChild("Gradient")
  local wndSprite = wndGradient:FindChild("Sprite")
  local wndMarker = wndGradient:FindChild("Marker")

  local nMarkerOffsetTop = math.max(0, math.min(1, nLastRelativeMouseY / wndSprite:GetHeight()))
  local nMarkerOffsetLeft = math.max(0, math.min(1, nLastRelativeMouseX / wndSprite:GetWidth()))

  wndMarker:SetAnchorPoints(nMarkerOffsetLeft, nMarkerOffsetTop, nMarkerOffsetLeft, nMarkerOffsetTop)
  self:ColorPickerChangeColor(self:ColorPickerReadFromMarkers())
end

-- ColorPicker:Alpha

function AlpacaColorPicker:OnColorPickerAlphaStart(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
  if wndHandler == wndControl then
    self.bColorPickerAlphaSelected = true
    self:ColorPickerMoveAlphaMarker(nLastRelativeMouseX, nLastRelativeMouseY)
  end
end

function AlpacaColorPicker:OnColorPickerAlphaStop(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation)
  if wndHandler == wndControl then
    self.bColorPickerAlphaSelected = false
  end
end

function AlpacaColorPicker:OnColorPickerAlphaMove(wndHandler, wndControl, nLastRelativeMouseX, nLastRelativeMouseY)
  if wndHandler == wndControl then
    if self.bColorPickerAlphaSelected then
      self:ColorPickerMoveAlphaMarker(nLastRelativeMouseX, nLastRelativeMouseY)
    end
  end
end

function AlpacaColorPicker:ColorPickerMoveAlphaMarker(nLastRelativeMouseX, nLastRelativeMouseY)
  local wndRange = self.wndColorPicker:FindChild("Alpha")
  local wndSprite = wndRange:FindChild("Sprite")
  local wndMarker = wndRange:FindChild("Marker")
  local nMarkerOffsetLeft = math.max(0, math.min(1, nLastRelativeMouseX / wndSprite:GetWidth()))
 
  wndMarker:SetAnchorPoints(nMarkerOffsetLeft, 0, nMarkerOffsetLeft, 1)
   
  self:ColorPickerChangeColor(self:ColorPickerReadFromMarkers())
end

-- ColorPicker:Numbers

function AlpacaColorPicker:OnColorPickerRedValueChanged(wndHandler, wndControl, strText)
  local nRed = tonumber(strText)

  if nRed and nRed >= 0 and nRed <= 255 then
    self:ColorPickerChangeColor(self:ColorPickerReadFromNumbers(), wndControl)
  end
end

function AlpacaColorPicker:OnColorPickerGreenValueChanged(wndHandler, wndControl, strText)
  local nGreen = tonumber(strText)

  if nGreen and nGreen >= 0 and nGreen <= 255 then
    self:ColorPickerChangeColor(self:ColorPickerReadFromNumbers(), wndControl)
  end
end

function AlpacaColorPicker:OnColorPickerBlueValueChanged(wndHandler, wndControl, strText)
  local nBlue = tonumber(strText)

  if nBlue and nBlue >= 0 and nBlue <= 255 then
    self:ColorPickerChangeColor(self:ColorPickerReadFromNumbers(), wndControl)
  end
end

function AlpacaColorPicker:OnColorPickerHexValueChanged(wndHandler, wndControl, strText)
  if string.len(strText) ~= 6 then
    return
  end

  local nRed = tonumber(string.sub(strText, 1, 2), 16)
  local nGreen = tonumber(string.sub(strText, 3, 4), 16)
  local nBlue = tonumber(string.sub(strText, 5, 6), 16)
  
  if nRed and nGreen and nBlue then
    local wndRedValue = self.wndColorPicker:FindChild("RedValue")
    local wndGreenValue = self.wndColorPicker:FindChild("GreenValue")
    local wndBlueValue = self.wndColorPicker:FindChild("BlueValue")

    wndRedValue:SetText(nRed)
    wndGreenValue:SetText(nGreen)
    wndBlueValue:SetText(nBlue)
          
    self:ColorPickerChangeColor(self:ColorPickerReadFromNumbers(), wndControl)
  end
end

function AlpacaColorPicker:OnColorPickerAlphaValueChanged(wndHandler, wndControl, strText)
  local fAlpha = tonumber(strText)
  
  if fAlpha and fAlpha >= 0 and fAlpha <= 1 then
    self:ColorPickerChangeColor(self:ColorPickerReadFromNumbers(), wndControl)
  end
end

-- ColorPicker:Buttons

function AlpacaColorPicker:OnColorPickerOkButtonSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler == wndControl then
    local tData = self.wndColorPicker:GetData()
    local crColor = self:ColorPickerReadFromPreview()
  
    if tData and tData.arSettingPath then    
      Event_FireGenericEvent("AlpacaColorPickerSelection", tData.objPlugin, tData.arSettingPath, crColor)
    end
      
    self.wndColorPicker:Show(false)
  end
end

function AlpacaColorPicker:OnColorPickerCancelButtonSignal(wndHandler, wndControl, eMouseButton)
  self.wndColorPicker:Show(false)
end

-- Start

local AlpacaColorPickerInst = AlpacaColorPicker:new()
AlpacaColorPickerInst:Init()

