require "Apollo"
require "ApolloColor" 
require "ApolloTimer"
require "GameLib"
require "XmlDoc"
require "Window"

local AlpacaPluginLibrary = nil
local AlpacaStandardLibrary = nil
local AlpacaWindowLibrary = nil

local knSettingsVersion = 2
local knProfileCount = 8
local knSliderMarkerSize = 18

local AlpacaHUD = {} 

function AlpacaHUD:Defaults()
  local tDefaults = {
    tAccount = {
      tDisplay = {
        tPosition = {
          fHorizontal = 0.5,
          fVertical = 0.5
        },
        tSize = {
          nWidth = 1000,
          nHeight = 1000
        }
      },
      arProfiles = {}
    },
    tCharacter = {
      nProfile = 1
    }
  }
  
  for i = 1, knProfileCount do
    table.insert(tDefaults.tAccount.arProfiles, {})
  end
  
  return tDefaults
end

function AlpacaHUD:Form()
  local fnCreateProfileOptions = function ()
    return {
      { strLabel = "Profile 1", nValue = 1 },
      { strLabel = "Profile 2", nValue = 2 },
      { strLabel = "Profile 3", nValue = 3 },
      { strLabel = "Profile 4", nValue = 4 },
      { strLabel = "Profile 5", nValue = 5 },
      { strLabel = "Profile 6", nValue = 6 },
      { strLabel = "Profile 7", nValue = 7 },
      { strLabel = "Profile 8", nValue = 8 }
    }
  end

  local arInputDefinitions = {
    AlpacaPluginLibrary:CreateHeader("Display"),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tAccount", "tDisplay", "tPosition", "fHorizontal" }, 0, 1, 0.01),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tAccount", "tDisplay", "tPosition", "fVertical" }, 0, 1, 0.01),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tAccount", "tDisplay", "tSize", "nWidth" }, 100, 3000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tAccount", "tDisplay", "tSize", "nHeight" }, 100, 3000, 1),
    AlpacaPluginLibrary:CreateHeader("Profile"),
    AlpacaPluginLibrary:CreateSelect("Current Profile",          { "tCharacter", "nProfile" }, fnCreateProfileOptions()),
  }
  
  return AlpacaPluginLibrary:CreateForm(arInputDefinitions)
end

function AlpacaHUD:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
    return o
end

function AlpacaHUD:Init()
  local bHasConfigureFunction = true
  local strConfigureButtonText = "AlpacaHUD"
  local arDependencies = { "AlpacaPluginLibrary", "AlpacaStandardLibrary", "AlpacaWindowLibrary" }
  
  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, arDependencies)
end

function AlpacaHUD:OnLoad()  
  Apollo.LoadSprites("AlpacaSprites.xml")
  
  AlpacaPluginLibrary = Apollo.GetPackage("AlpacaPluginLibrary").tPackage
  AlpacaStandardLibrary = Apollo.GetPackage("AlpacaStandardLibrary").tPackage
  AlpacaWindowLibrary = Apollo.GetPackage("AlpacaWindowLibrary").tPackage

  self.arPlugins = {}
  self.tSettings = self:Defaults()

  self.xmlDoc = XmlDoc.CreateFromFile("AlpacaControlPanel.xml")
  self.timer = ApolloTimer.Create(0.1, false, "OnTimer", self)
end

function AlpacaHUD:IsReady() 
  if not self.bReady then
    if not AlpacaWindowLibrary:IsLoaded() then 
      return false
    end
    
    if not self.xmlDoc then
      return false
    end
    
    self.bReady = true
  end
  
  return self.bReady
end

function AlpacaHUD:OnTimer()
  if not self:IsReady() then
    self.timer = ApolloTimer.Create(0.1, false, "OnTimer", self)
    return
  end
 
  self:CreateDisplay()
  
  Apollo.RegisterEventHandler("AlpacaColorPickerSelection", "OnAlpacaColorPickerSelection", self)
  Apollo.RegisterEventHandler("AlpacaRegisterPlugin", "OnAlpacaRegisterPlugin", self)
  
  Apollo.RegisterSlashCommand("alpaca", "OnSlashCommand", self)  
end

function AlpacaHUD:OnAlpacaRegisterPlugin(objPlugin)
  local bExists = false
  
  for _, objExisting in ipairs(self.arPlugins) do
    if objExisting == objPlugin then
      bExists = true
    end
  end
  
  if not bExists then
    table.insert(self.arPlugins, objPlugin)

    local fnCompareByPluginName = function (a, b)
      return string.lower(a:PluginName()) < string.lower(b:PluginName())
    end
      
    table.sort(self.arPlugins, fnCompareByPluginName)

    objPlugin:OnPluginRegistered(self.wndDisplay)
  end
  
  self:SyncPluginSettings(objPlugin)
  
  Event_FireGenericEvent("AlpacaRegisterPluginSuccess", objPlugin)
end

function AlpacaHUD:OnSlashCommand(strCommand, strParam)
  if strParam == "debug" then
    Event_FireGenericEvent("AlpacaOpenDebug")
  else
    self:OpenControlPanel()
  end	
end

-- Settings

function AlpacaHUD:OnSave(eLevel)
  if eLevel == GameLib.CodeEnumAddonSaveLevel.Account then 
    return { nSettingsVersion = knSettingsVersion, tSettings = self.tSettings.tAccount }
  end  
  
  if eLevel == GameLib.CodeEnumAddonSaveLevel.Character then 
    return { nSettingsVersion = knSettingsVersion, tSettings = self.tSettings.tCharacter }
  end
end

function AlpacaHUD:OnRestore(eLevel, tData)   
  if not tData or tData.nSettingsVersion ~= knSettingsVersion then
    return
  end
  
  if eLevel == GameLib.CodeEnumAddonSaveLevel.Account then 
    self:SyncAccountSettings(tData.tSettings)
  end
  
  if eLevel == GameLib.CodeEnumAddonSaveLevel.Character then 
    self:SyncCharacterSettings(tData.tSettings)
  end
end

function AlpacaHUD:SyncAccountSettings(tAccountSettings)
  self.tSettings.tAccount.tDisplay = self:SyncSettings(self:Defaults().tAccount.tDisplay, (tAccountSettings and tAccountSettings.tDisplay))  
 
  self.tSettings.tAccount.arProfiles = {}
  
  for i = 1, knProfileCount do  
    local tProfile = {}
    
    if tAccountSettings and tAccountSettings.arProfiles and tAccountSettings.arProfiles[i] then
      tProfile = tAccountSettings.arProfiles[i]
    end
    
    self.tSettings.tAccount.arProfiles[i] = tProfile
  end
  
  for _, objPlugin in ipairs(self.arPlugins) do
    self:SyncPluginSettings(objPlugin)
  end
end

function AlpacaHUD:SyncCharacterSettings(tCharacterSettings)
  self.tSettings.tCharacter = self:SyncSettings(self:Defaults().tCharacter, tCharacterSettings)
end

function AlpacaHUD:SyncPluginSettings(objPlugin)
  local strPluginId = objPlugin:PluginId()
  local tDefaults = objPlugin:PluginDefaults()
  
  for i = 1, #self.tSettings.tAccount.arProfiles do    
    local tProfile = self.tSettings.tAccount.arProfiles[i]
    
    tProfile[strPluginId] = self:SyncSettings(tDefaults, tProfile[strPluginId])
  end
  
  objPlugin:OnPluginSettingsChanged(self:GetPluginSettings(objPlugin))
  
  self:RefreshPluginPages(objPlugin)
end

function AlpacaHUD:SyncSettings(tDefaults, tCurrent) 
  if tDefaults == nil or type(tDefaults) ~= type(tCurrent) then
    return AlpacaStandardLibrary:Copy(tDefaults)
  elseif type(tDefaults) == "table" and type(tCurrent) == "table" then
    local tMerge = {}
    for key, _ in pairs(tDefaults) do
      tMerge[key] = self:SyncSettings(tDefaults[key], tCurrent[key])
    end
    return tMerge
  else  
    local xDefaults = AlpacaStandardLibrary:Copy(tDefaults)
    local xCurrent = AlpacaStandardLibrary:Copy(tCurrent)
  
    if xCurrent ~= nil then
      return xCurrent 
    else
      return xDefaults
    end
  end
end

function AlpacaHUD:GetAddonSetting(arSettingPath)
  return self:GetSetting(self.tSettings, arSettingPath)
end

function AlpacaHUD:SetAddonSetting(arSettingPath, xValue)
  if self:SetSetting(self.tSettings, arSettingPath, xValue) then
    if arSettingPath[1] == "tCharacter" and arSettingPath[2] == "nProfile" then
      self:EnablePreview(false)
      
      for _, objPlugin in ipairs(self.arPlugins) do
        objPlugin:OnPluginSettingsChanged(self:GetPluginSettings(objPlugin))
      end
    else
      self:RefreshDisplay()
    end
  end
end

function AlpacaHUD:GetPluginSettings(objPlugin)
  return self.tSettings.tAccount.arProfiles[self.tSettings.tCharacter.nProfile][objPlugin:PluginId()]
end

function AlpacaHUD:GetPluginSetting(objPlugin, arSettingPath)
  local tSettings = self:GetPluginSettings(objPlugin)
  return self:GetSetting(tSettings, arSettingPath)
end

function AlpacaHUD:SetPluginSetting(objPlugin, arSettingPath, xValue)
  local tSettings = self:GetPluginSettings(objPlugin)
  
  if self:SetSetting(tSettings, arSettingPath, xValue) then
    if objPlugin.OnPluginSettingsChanged then
      objPlugin:OnPluginSettingsChanged(tSettings)
    end
  end
end

function AlpacaHUD:GetSetting(tSettings, arSettingPath)  
  local tCurrentNode = tSettings
  for i, key in ipairs(arSettingPath) do
    if i < #arSettingPath then
      if not tCurrentNode[key] then 
        tCurrentNode[key] = {} 
      end
      tCurrentNode = tCurrentNode[key]  
    else
      return tCurrentNode[key]
    end
  end
end

function AlpacaHUD:SetSetting(tSettings, arSettingPath, xValue)  
  local tCurrentNode = tSettings
  for i, key in ipairs(arSettingPath) do
    if i < #arSettingPath then
      if not tCurrentNode[key] then 
        tCurrentNode[key] = {} 
      end
      tCurrentNode = tCurrentNode[key]
    else
      if tCurrentNode[key] ~= xValue then
        tCurrentNode[key] = xValue
        return true
      end  
    end
  end
end

function AlpacaHUD:EnablePreview(bEnable) 
  if self.wndControlPanel then
    self.wndControlPanel:FindChild("PreviewButton"):SetCheck(bEnable)
  end
  
  for _, objPlugin in ipairs(self.arPlugins) do
    self:SetPluginSetting(objPlugin, { "bPreview" }, bEnable)
  end
  
  self:SetAddonSetting({ "bPreview" }, bEnable)
end

-- Display

function AlpacaHUD:CreateDisplay()
  if self.wndDisplay then
    return
  end
  
  self.wndDisplay = AlpacaWindowLibrary:CreateEmptyFrame()
  
  self:RefreshDisplay()
end

function AlpacaHUD:RefreshDisplay()
  if not self.wndDisplay then
    return
  end
  
  local bPreview = self.tSettings.bPreview
  
  local tDisplaySettings = self.tSettings.tAccount.tDisplay
  local fHorizontal = tDisplaySettings.tPosition.fHorizontal
  local fVertical = tDisplaySettings.tPosition.fVertical
  local nWidth = tDisplaySettings.tSize.nWidth
  local nHeight = tDisplaySettings.tSize.nHeight
  local nLeft = -1 * math.floor(nWidth / 2)
  local nTop = -1 * math.floor(nHeight / 2)
  local nRight = nLeft + nWidth
  local nBottom = nTop + nHeight
  
  self.wndDisplay:SetAnchorPoints(fHorizontal, fVertical, fHorizontal, fVertical)
  self.wndDisplay:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)
  
  if bPreview then 
    self.wndDisplay:SetSprite("AlpacaSprites:WhiteFill")
    self.wndDisplay:SetBGColor(ApolloColor.new(1, 1, 1, 0.25))
  else
    self.wndDisplay:SetSprite(nil)
  end
end

-- Control Panel Events

function AlpacaHUD:OnWindowClosed(wndHandler, wndControl)
	if wndHandler == wndControl then
		self:CloseControlPanel()
	end
end

function AlpacaHUD:OnCloseButtonSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler == wndControl then
    self:CloseControlPanel()
  end
end

function AlpacaHUD:OnPreviewButtonSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler == wndControl then    
    self:EnablePreview(wndControl:IsChecked())
  end
end

function AlpacaHUD:OnTabButtonSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler == wndControl then
    self:SetCurrentPage(wndHandler:GetData().strId)
  end
end

function AlpacaHUD:OnWindowMouseButtonDown(wndHandler, wndControl)
  for _, wndPage in ipairs(wndHandler:FindChild("Pages"):GetChildren()) do
    for _, wndInput in ipairs(wndPage:GetChildren()) do
      local wndEditBox = wndInput:FindChild("EditBox")
      
      if wndEditBox then
        wndEditBox:ClearFocus()
      end
    end
  end
end

-- Control Panel

function AlpacaHUD:OpenControlPanel()
  self:CloseControlPanel()
  
  self.wndControlPanel = Apollo.LoadForm(self.xmlDoc, "ControlPanel", nil, self)
  self.wndControlPanel:Show(false, true)
  self.wndControlPanel:SetSizingMinimum(600, 600)
  
  self:CreateTabs()
  self:CreatePages()
  self:SetCurrentPage("Addon")
  
  self.wndControlPanel:Show(true, true)
end

function AlpacaHUD:CloseControlPanel()  
  self:EnablePreview(false)
  
  if self.wndControlPanel then   
    self.wndControlPanel:Destroy()
    self.wndControlPanel = nil
  end
end

function AlpacaHUD:GetCurrentPage()
  for _, wndPage in ipairs(self.wndControlPanel:FindChild("Pages"):GetChildren()) do
    if wndPage:IsShown() then
      return wndPage:GetData().strId
    end
  end
end

function AlpacaHUD:SetCurrentPage(strId)
  for _, wndTab in ipairs(self.wndControlPanel:FindChild("Tabs"):GetChildren()) do
    wndTab:SetCheck(wndTab:GetData().strId == strId)
  end
  
  for _, wndPage in ipairs(self.wndControlPanel:FindChild("Pages"):GetChildren()) do
    wndPage:Show(wndPage:GetData().strId == strId)    
  end
end

-- Tabs

function AlpacaHUD:CreateTabs() 
  local wndTabs = self.wndControlPanel:FindChild("Tabs") 
  
  for _, wndTab in ipairs(wndTabs:GetChildren()) do
     wndTab:Destroy()  
  end
  
  self:CreateAddonTab()
  
  for _, objPlugin in ipairs(self.arPlugins) do
    self:CreatePluginTabs(objPlugin)
  end  
end

function AlpacaHUD:CreateAddonTab()
  local wndTabs = self.wndControlPanel:FindChild("Tabs") 
  local wndTab = Apollo.LoadForm(self.xmlDoc, "Tab", wndTabs, self)
    
  wndTab:SetText("General")
  wndTab:SetData({ strId = "Addon" })
  wndTab:Show(true, true)    
end

function AlpacaHUD:CreatePluginTabs(objPlugin)
  if not objPlugin.PluginForm then
    return
  end

  local wndTabs = self.wndControlPanel:FindChild("Tabs")
  local tForm = objPlugin:PluginForm()

  local fnCreateTab = function (objPlugin, tSubForm)
    local wndTab = Apollo.LoadForm(self.xmlDoc, "Tab", wndTabs, self)
    local strLabel = tSubForm and tSubForm.strLabel or objPlugin:PluginName()
    
    wndTab:SetText((tSubForm and "    " or "") .. strLabel)
    wndTab:SetData({ objPlugin = objPlugin, strId = objPlugin:PluginId() .. ":" .. strLabel })  
    wndTab:Show(true, true)
    
    if tSubForm then
      local crWhite = ApolloColor.new(1, 1, 1, 1)
      
      wndTab:SetFont("DefaultButton")
      wndTab:SetNormalTextColor(crWhite)
      wndTab:SetPressedTextColor(crWhite)
      wndTab:SetFlybyTextColor(crWhite)
      wndTab:SetPressedFlybyTextColor(crWhite)
      wndTab:SetDisabledTextColor(crWhite)
    end
  end
    
  fnCreateTab(objPlugin)
  
  if tForm.arSubForms then 
    for _, tSubForm in ipairs(tForm.arSubForms) do
      fnCreateTab(objPlugin, tSubForm)
    end
  end
 
  wndTabs:ArrangeChildrenVert(0)  
end

-- Pages

function AlpacaHUD:CreatePages()
  local wndPages = self.wndControlPanel:FindChild("Pages")
  
  for _, wndPage in ipairs(wndPages:GetChildren()) do
    wndPage:Destroy()  
  end
  
  self:CreateAddonPage()
  
  for _, objPlugin in ipairs(self.arPlugins) do
    self:CreatePluginPages(objPlugin)
  end    
end

function AlpacaHUD:CreateAddonPage()
  local wndPages = self.wndControlPanel:FindChild("Pages")
  local wndPage = Apollo.LoadForm(self.xmlDoc, "Page", wndPages, self)
  local tForm = self:Form()
  
  wndPage:SetData({ strId = "Addon" })  
  wndPage:Show(false, true)
    
  for _, tInputDefinition in ipairs(tForm.arInputDefinitions) do
    self:CreateInput(wndPage, tInputDefinition)
  end
  
  wndPage:ArrangeChildrenVert(0)
  
  self:RefreshAddonPage()
end

function AlpacaHUD:RefreshAddonPage()
  if not self.wndControlPanel then
    return
  end
  
  for _, wndPage in ipairs(self.wndControlPanel:FindChild("Pages"):GetChildren()) do
    if wndPage:GetData().strId == "Addon" then
      for _, wndInput in ipairs(wndPage:GetChildren()) do
        self:RefreshInput(wndInput)
      end
    end
  end
end

function AlpacaHUD:IsAddonPage(wndPage)
  return wndPage:GetData().strId == "Addon"
end

function AlpacaHUD:CreatePluginPages(objPlugin)  
  if not objPlugin.PluginForm then
    return
  end

  local wndPages = self.wndControlPanel:FindChild("Pages")
  local tForm = objPlugin:PluginForm()

  local fnCreatePage = function (objPlugin, tSubForm)
    local wndPage = Apollo.LoadForm(self.xmlDoc, "Page", wndPages, self)
    local strLabel = tSubForm and tSubForm.strLabel or objPlugin:PluginName()

    wndPage:SetData({ objPlugin = objPlugin, strId = objPlugin:PluginId() .. ":" .. strLabel })  
    wndPage:Show(false, true)
    
    local arInputDefinitions = tSubForm and tSubForm.arInputDefinitions or tForm.arInputDefinitions

    for _, tInputDefinition in ipairs(arInputDefinitions) do
      self:CreateInput(wndPage, tInputDefinition)
    end
  
    wndPage:ArrangeChildrenVert(0)
  end

  fnCreatePage(objPlugin)
  
  if tForm.arSubForms then 
    for _, tSubForm in ipairs(tForm.arSubForms) do
      fnCreatePage(objPlugin, tSubForm)
    end
  end

  self:RefreshPluginPages(objPlugin)
end

function AlpacaHUD:RefreshPluginPages(objPlugin) 
  if not self.wndControlPanel then
    return
  end
  
  for _, wndPage in ipairs(self.wndControlPanel:FindChild("Pages"):GetChildren()) do
    if wndPage:GetData().objPlugin == objPlugin then
      for _, wndInput in ipairs(wndPage:GetChildren()) do
        self:RefreshInput(wndInput)
      end
    end
  end
end

-- Inputs

function AlpacaHUD:CreateInput(wndPage, tInputDefinition) 
  local tFrameLookup = {
    [AlpacaPluginLibrary.EnumInputType.Header]        = "Header",
    [AlpacaPluginLibrary.EnumInputType.SubHeader]     = "SubHeader",
    [AlpacaPluginLibrary.EnumInputType.Checkbox]      = "Checkbox",
    [AlpacaPluginLibrary.EnumInputType.Color]         = "Color",
    [AlpacaPluginLibrary.EnumInputType.Select]        = "Select",
    [AlpacaPluginLibrary.EnumInputType.Slider]        = "Slider"
  }
  
  local strFrameName = tFrameLookup[tInputDefinition.eInputType]
  
  if not strFrameName then
    return   
  end
  
  local wndInput = Apollo.LoadForm(self.xmlDoc, strFrameName, wndPage, self)
  
  wndInput:SetText(tInputDefinition.strLabel)
  wndInput:SetData(tInputDefinition)  
        
  if tInputDefinition.eInputType == AlpacaPluginLibrary.EnumInputType.Checkbox then
    self:InitializeCheckbox(wndInput)
  elseif tInputDefinition.eInputType == AlpacaPluginLibrary.EnumInputType.Color then
    self:InitializeColor(wndInput)
  elseif tInputDefinition.eInputType == AlpacaPluginLibrary.EnumInputType.Select then
    self:InitializeSelect(wndInput)
  elseif tInputDefinition.eInputType == AlpacaPluginLibrary.EnumInputType.Slider then
    self:InitializeSlider(wndInput)
  end
end

function AlpacaHUD:RefreshInput(wndInput)
  local eInputType = wndInput:GetData().eInputType
  
  if eInputType == AlpacaPluginLibrary.EnumInputType.Checkbox then
    self:RefreshCheckbox(wndInput)
  elseif eInputType == AlpacaPluginLibrary.EnumInputType.Color then
    self:RefreshColor(wndInput) 
  elseif eInputType == AlpacaPluginLibrary.EnumInputType.Select then
    self:RefreshSelect(wndInput)
  elseif eInputType == AlpacaPluginLibrary.EnumInputType.Slider then
    self:RefreshSlider(wndInput)  
  end
end

-- Checkbox

function AlpacaHUD:InitializeCheckbox(wndInput)

end

function AlpacaHUD:RefreshCheckbox(wndInput)
  local wndPage = AlpacaWindowLibrary:FindParent(wndInput, "Page")
  local wndButton = wndInput:FindChild("Button")
  local tPageData = wndPage:GetData()
  local tInputData = wndInput:GetData()
  local bValue = nil
  
  if self:IsAddonPage(wndPage) then
    bValue = self:GetAddonSetting(tInputData.arSettingPath)
  else
    bValue = self:GetPluginSetting(tPageData.objPlugin, tInputData.arSettingPath)
  end
  
  wndButton:SetCheck(bValue == true)  
end

function AlpacaHUD:OnCheckboxButtonSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler ~= wndControl then
    return
  end
  
  local wndPage = AlpacaWindowLibrary:FindParent(wndControl, "Page")
  local wndInput = AlpacaWindowLibrary:FindParent(wndControl, "Checkbox")
  local tPageData = wndPage:GetData()
  local tInputData = wndInput:GetData()
  local bValue = wndHandler:IsChecked()
    
  if self:IsAddonPage(wndPage) then
    self:SetAddonSetting(tInputData.arSettingPath, bValue)
  else
    self:SetPluginSetting(tPageData.objPlugin, tInputData.arSettingPath, bValue)
  end
end

-- Color

function AlpacaHUD:InitializeColor(wndInput)

end

function AlpacaHUD:RefreshColor(wndInput)
  local wndPage = AlpacaWindowLibrary:FindParent(wndInput, "Page")
  local wndButton = wndInput:FindChild("Button")  
  local tPageData = wndPage:GetData()
  local tInputData = wndInput:GetData()
  local crColor = nil
  
  if self:IsAddonPage(wndPage) then
    crColor = self:GetAddonSetting(tInputData.arSettingPath)
  else
    crColor = self:GetPluginSetting(tPageData.objPlugin, tInputData.arSettingPath)
  end
    
  wndButton:SetBGColor(AlpacaWindowLibrary:ToApolloColor(crColor))  
end

function AlpacaHUD:OnColorButtonSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler ~= wndControl then
    return
  end
  
  local wndInput = AlpacaWindowLibrary:FindParent(wndHandler, "Color")
  local wndPage = AlpacaWindowLibrary:FindParent(wndHandler, "Page")
  local tPageData = wndPage:GetData()
  local tInputData = wndInput:GetData()
  local crColor = nil
  
  if self:IsAddonPage(wndPage) then
    crColor = self:GetAddonSetting(tInputData.arSettingPath)
  else
    crColor = self:GetPluginSetting(tPageData.objPlugin, tInputData.arSettingPath)
  end
    
  Event_FireGenericEvent("OpenAlpacaColorPicker", wndInput, tInputData.arSettingPath, crColor)
end

function AlpacaHUD:OnAlpacaColorPickerSelection(wndInput, arSettingPath, crColor)
  local wndPage = AlpacaWindowLibrary:FindParent(wndInput, "Page")
  local tPageData = wndPage:GetData()
  local tInputData = wndInput:GetData()
  
  if self:IsAddonPage(wndPage) then
    self:SetAddonSetting(tInputData.arSettingPath, crColor)
  else
    self:SetPluginSetting(tPageData.objPlugin, tInputData.arSettingPath, crColor)
  end
  
  self:RefreshColor(wndInput)
end

-- Select

function AlpacaHUD:InitializeSelect(wndInput)
  local tInputData = wndInput:GetData()
  local wndPopUp = wndInput:FindChild("PopUp")
  local wndOptions = wndPopUp:FindChild("Options")
  
  for i, wndOption in ipairs(wndOptions:GetChildren()) do
    local tOption = tInputData.arOptions[i]
    
    if tOption then
      wndOption:SetText(tOption.strLabel)
      wndOption:SetData(tOption.nValue)
    else
      wndOption:Destroy()
    end    
  end
      
  local nOffsetLeft = 31
  local nOffsetTop = 8
      
  wndPopUp:SetAnchorPoints(1, 0, 1, 0)
  wndPopUp:SetAnchorOffsets(nOffsetLeft, nOffsetTop, nOffsetLeft + 200, nOffsetTop + wndOptions:ArrangeChildrenVert(0) + 10)
end

function AlpacaHUD:RefreshSelect(wndInput)
  local wndPage = AlpacaWindowLibrary:FindParent(wndInput, "Page")
  local wndButton = wndInput:FindChild("Button")
  local wndPopUp = wndInput:FindChild("PopUp")
  local wndOptions = wndInput:FindChild("Options")
  local tPageData = wndPage:GetData()
  local tInputData = wndInput:GetData()
  local nValue = nil
    
  if self:IsAddonPage(wndPage) then
    nValue = self:GetAddonSetting(tInputData.arSettingPath)
  else
    nValue = self:GetPluginSetting(tPageData.objPlugin, tInputData.arSettingPath)
  end
  
  wndButton:SetCheck(false)
  wndPopUp:Show(false, true)
  
  for _, wndOption in ipairs(wndOptions:GetChildren()) do
    if wndOption:GetData() == nValue then
      wndButton:SetText(wndOption:GetText())
      wndOption:SetCheck(true)
    else
      wndOption:SetCheck(false)
    end
  end
end

function AlpacaHUD:OnSelectButtonSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler ~= wndControl then
    return
  end
  
  local wndPage = AlpacaWindowLibrary:FindParent(wndHandler, "Page")
  local wndPopUp = wndHandler:FindChild("PopUp")
  local tPageData = wndPage:GetData()
  local bIsChecked = wndHandler:IsChecked()

  if self:IsAddonPage(wndPage) then
    self:RefreshAddonPage()
  else
    self:RefreshPluginPages(tPageData.objPlugin)
  end
 
  wndHandler:SetCheck(bIsChecked)
  wndPopUp:Show(bIsChecked, true)
end

function AlpacaHUD:OnSelectOptionSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler ~= wndControl then
    return
  end
    
  local wndPage = AlpacaWindowLibrary:FindParent(wndHandler, "Page")
  local wndInput = AlpacaWindowLibrary:FindParent(wndHandler, "Select")
  local tPageData = wndPage:GetData()
  local tInputData = wndInput:GetData()
  local nValue = wndHandler:GetData()
      
  if self:IsAddonPage(wndPage) then
    self:SetAddonSetting(tInputData.arSettingPath, nValue)
    self:RefreshAddonPage()
  else
    self:SetPluginSetting(tPageData.objPlugin, tInputData.arSettingPath, nValue)
    self:RefreshPluginPages(tPageData.objPlugin)
  end
end

-- Slider

function AlpacaHUD:InitializeSlider(wndInput)

end

function AlpacaHUD:RefreshSlider(wndInput)
  local wndPage = AlpacaWindowLibrary:FindParent(wndInput, "Page")
  local tPageData = wndPage:GetData()
  local tInputData = wndInput:GetData()
  local fValue = nil
  
  if self:IsAddonPage(wndPage) then
    fValue = self:GetAddonSetting(tInputData.arSettingPath)
  else
    fValue = self:GetPluginSetting(tPageData.objPlugin, tInputData.arSettingPath)
  end

  self:ChangeSliderValue(wndInput, fValue)
end

function AlpacaHUD:OnSliderLeftButtonSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler ~= wndControl then
    return
  end
    
  local wndInput = AlpacaWindowLibrary:FindParent(wndHandler, "Slider")
  local tData = wndInput:GetData()
  
  if tData.fValue - tData.fStep >= tData.fMin then
    self:ChangeSliderValue(wndInput, tData.fValue - tData.fStep)
  end
end

function AlpacaHUD:OnSliderRightButtonSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler ~= wndControl then
    return
  end

  local wndInput = AlpacaWindowLibrary:FindParent(wndHandler, "Slider")
  local tData = wndInput:GetData()
  
  if tData.fValue + tData.fStep <= tData.fMax then        
    self:ChangeSliderValue(wndInput, tData.fValue + tData.fStep)
  end
end

function AlpacaHUD:OnSliderMarkerStart(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
  if wndHandler ~= wndControl then
    return
  end
  
  local wndInput = AlpacaWindowLibrary:FindParent(wndHandler, "Slider")
    
  wndInput:GetData().bMarkerActive = true

  self:OnSliderMarkerMove(wndHandler, wndControl, nLastRelativeMouseX, nLastRelativeMouseY)
end

function AlpacaHUD:OnSliderMarkerStop(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation)
  if wndHandler ~= wndControl then
    return
  end
  
  local wndInput = AlpacaWindowLibrary:FindParent(wndHandler, "Slider")
    
  wndInput:GetData().bMarkerActive = false
end

function AlpacaHUD:OnSliderMarkerMove(wndHandler, wndControl, nLastRelativeMouseX, nLastRelativeMouseY)
  if wndHandler ~= wndControl then
    return
  end
    
  local wndInput = AlpacaWindowLibrary:FindParent(wndHandler, "Slider")
  local tData = wndInput:GetData()

  if tData.bMarkerActive then
    local nSliderLength = wndHandler:GetWidth() - knSliderMarkerSize
    local fPercent = math.max(0, math.min(nSliderLength, nLastRelativeMouseX - knSliderMarkerSize / 2)) / nSliderLength
    local fValue = fPercent * (tData.fMax - tData.fMin) + tData.fMin      
    local fRoundedValue = math.floor(fValue * (1 / tData.fStep)) / (1 / tData.fStep)
  
    self:ChangeSliderValue(wndInput, fRoundedValue)    
  end
end

function AlpacaHUD:OnSliderEditBoxChanged(wndHandler, wndControl, strText)
  if wndHandler ~= wndControl then
    return
  end
  
  local wndInput = AlpacaWindowLibrary:FindParent(wndHandler, "Slider")

  self:ChangeSliderValue(wndInput, tonumber(strText), wndControl)
end

function AlpacaHUD:ChangeSliderValue(wndInput, fValue, wndControl)
  local wndPage = AlpacaWindowLibrary:FindParent(wndInput, "Page")
  local wndMarkerRange = wndInput:FindChild("MarkerRange")
  local wndMarker = wndInput:FindChild("Marker")
  local wndEditBox = wndInput:FindChild("EditBox")
  local tPageData = wndPage:GetData()
  local tInputData = wndInput:GetData()
 
  if fValue and fValue >= tInputData.fMin and fValue <= tInputData.fMax then
    local fPercent = fValue / (tInputData.fMax - tInputData.fMin)
    local nSliderLength = wndMarkerRange:GetWidth() - knSliderMarkerSize
    local nMid = fPercent * nSliderLength + knSliderMarkerSize / 2
    local nLeft = math.ceil(nMid - knSliderMarkerSize / 2)
    local nRight = math.floor(nMid + knSliderMarkerSize / 2)   
    
    wndMarker:SetAnchorOffsets(nLeft, -9, nRight, 9)
    
    if wndControl ~= wndEditBox then
      wndEditBox:SetText(fValue)
    end

    wndEditBox:SetTextColor("UI_TextHoloBody")
        
    tInputData.fValue = fValue 
     
    if self:IsAddonPage(wndPage) then
      self:SetAddonSetting(tInputData.arSettingPath, fValue)
    else
      self:SetPluginSetting(tPageData.objPlugin, tInputData.arSettingPath, fValue)
    end
  else
    wndEditBox:SetTextColor(ApolloColor.new(1, 0, 0, 1))
  end  
end

-- Init

AlpacaHUD:new():Init()